/*******************************************************************************
  File: BtreeIndex.hxx
  Copyright (c) 2002 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, such as copyrighting
  or claiming authorship, to prevent the continued distribution of this 
  Software under an open source distribution model.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
	The template implementation file for the BtreeIndex class.
  
//#----------------------------------------------------------------------------|
Notes:
  
//%****************************************************************************/

// My header
#include "IndexedFile.h"

namespace Util {


//
// BtreeIndex functions
//
template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::BtreeIndex(Compare comp = Compare())
  : _comp(comp),
    _xstream(),
	_root(null),
	_pageCache(_xstream),
	_dStrategy(Compact),
	_uStrategy(MostlySafe)
{
	// empty
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::BtreeIndex(
			const std::string& name, 
			std::ios_base::openmode mode = std::ios_base::in|std::ios_base::out,
			Compare comp = Compare())
  : _comp(comp),
    _xstream(),
	_root(null),
	_pageCache(_xstream),
	_dStrategy(Compact),
	_uStrategy(MostlySafe)
{
	open(name, mode);
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::~BtreeIndex()
{
	close();
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
void
BtreeIndex<Key,Compare,KeyTraits,PageSize>::open(
			const std::string& name, 
			std::ios_base::openmode mode = std::ios_base::in|std::ios_base::out)
{
	const char* const func = "BtreeIndex::open";

	// the index file can not be opened just for output,
	// so if the user specifies 'out' (to create the file)
	// we first open it, then close it and re-open it.
	if ( not (mode & std::ios_base::in) ) {
		_xstream.open(name, mode);
		_xstream.close();
		_xstream.open(name, mode | std::ios_base::in);
	} else {
		_xstream.open(name, mode);
	}
	if (not _xstream.good()) raise(
		std::ios_base::failure(
			make_what_str("ios_base::failure",
				std::string("Unable to open BtreeIndex XDRStream file ").append(name).append(" for the mode specified"),
				func)));

	// read in the page size, part of the first element in the root page
	unsigned long temp = 0;
	_xstream >> temp;
	size_t pageSize = temp & 0x0000ffff;

	if (_xstream.eof()) {
		// if we didn't get anything at all, create a new root page
		_xstream.clear();
		_root = new RootPage;
		_root->encode(_xstream);
	} else if (pageSize != PageSize) {	
		// if the stored page size does not match the class' throw exception
		raise(std::ios_base::failure(
			make_what_str("ios::failure", 
				"Page size of file does not match template argument of class", func)));
	} else {
		// load the root page
		_xstream.seekg(0);	// return to the beginning
		_root = new RootPage;
		_root->decode(_xstream);
		_pageCache.min_size(_root->depth());
	}
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
void
BtreeIndex<Key,Compare,KeyTraits,PageSize>::close()
{
	if (not is_open()) return;
	flush();			// write out all dirty pages
	_pageCache.clear();	// delete all cached pages
	delete _root;		// ditto for the root
	_xstream.close();
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
void
BtreeIndex<Key,Compare,KeyTraits,PageSize>::flush()
{
	flush(Safe);
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::iterator
BtreeIndex<Key,Compare,KeyTraits,PageSize>::begin()
{
	if (empty()) return end();
	return lower_bound( _root->key_list().front().key );
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::iterator
BtreeIndex<Key,Compare,KeyTraits,PageSize>::end()
{
	return Iterator(this, -1, 0);
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::iterator
BtreeIndex<Key,Compare,KeyTraits,PageSize>::insert(const value_type& x)
{
	Iterator it = insert(_root, _root->depth(), x);
	if (_root->must_split()) {
		_root->split(*this);
		_pageCache.min_size(_root->depth());
	}
	_root->incr_total();
	flush(_uStrategy);
	return it;
}


#ifndef NESTED_TEMPLATE_DEFINITION_NOT_SUPPORTED
template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
template<typename InputIterator>
void
BtreeIndex<Key,Compare,KeyTraits,PageSize>::insert(InputIterator first, InputIterator last)
{
	while (first != last) {
		insert(_root, _root->depth, *first);
		if (_root->must_split()) _root->split();
		_root->incr_total();
		++first;
	}
	flush(_uStrategy);
}
#endif


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
void
BtreeIndex<Key,Compare,KeyTraits,PageSize>::erase(Iterator pos)
{
	// The 'pos' does not contain the key, but we need it to chase the links
	Key k = (*pos).first;

	// Find a path to the leaf page and remove the key
	erase(_root, _root->depth(), k, pos);
	_root->decr_total();
	// See if the changes perculated to the root
	if (_root->depth() > 0 and _root->num_keys() == 1) {
		_root->consolidate(*this);
		_pageCache.min_size(_root->depth());
	} else if (_root->must_split()) {
		_root->split(*this);
		_pageCache.min_size(_root->depth());
	}
	flush(_uStrategy);
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::size_type
BtreeIndex<Key,Compare,KeyTraits,PageSize>::erase(const Key& x)
{
	// TBD
	return 0;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
void
BtreeIndex<Key,Compare,KeyTraits,PageSize>::erase(iterator first, Iterator last)
{ 
	// TBD
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
void
BtreeIndex<Key,Compare,KeyTraits,PageSize>::clear()
{
	// TBD
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::iterator
BtreeIndex<Key,Compare,KeyTraits,PageSize>::find(const Key& k)
{
	// Use the internal lower_bound
	if (empty()) return end();
	Iterator it = lower_bound(_root, _root->depth(), k);
	if (it != end() and _comp(k, (*it).first)) it = end();
	_pageCache.reset();
	return it;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::size_type
BtreeIndex<Key,Compare,KeyTraits,PageSize>::count(const Key& x)
{
	// TBD
	return 0;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::iterator
BtreeIndex<Key,Compare,KeyTraits,PageSize>::lower_bound(const Key& k)
{
	if (empty()) return end();
	Iterator it = lower_bound(_root, _root->depth(), k);
	_pageCache.reset();
	return it;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::iterator
BtreeIndex<Key,Compare,KeyTraits,PageSize>::upper_bound(const Key& k)
{
	if (empty()) return end();
	Iterator it = upper_bound(_root, _root->depth(), k);
	_pageCache.reset();
	return it;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
std::pair<BtreeIndex<Key,Compare,KeyTraits,PageSize>::iterator, 
          BtreeIndex<Key,Compare,KeyTraits,PageSize>::iterator>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::equal_range(const Key& x)
{
	return std::make_pair(lower_bound(x), upper_bound(x));
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
void
BtreeIndex<Key,Compare,KeyTraits,PageSize>::reindex()
{
	// TBD
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
std::pair<const Key, BtreeIndex<Key,Compare,KeyTraits,PageSize>::StreamposProxy>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::get_value(size_t page, size_t slot)
{
	Page* p = get_page(page);
	// Find the slot
	Page::KeyList::iterator kit = p->key_list().begin();
	while (slot > 0 and kit != p->key_list().end()) {
		++kit;
		--slot;
	}
	if (kit == p->key_list().end()) throw std::ios_base::failure("Invalid slot");
	return std::make_pair((*kit).key, StreamposProxy((*kit).link, this, page, slot));
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
void
BtreeIndex<Key,Compare,KeyTraits,PageSize>::prev_iter(Iterator& it)
{
	if (it._slot > 0) --it._slot;
	else {
		Page* p = get_page(it._page);
		Page* prevp = get_page(p->prev_page());
		if (prevp) {
			it._page = prevp->page_id();
			it._slot = prevp->num_keys() - 1;
		}
	}
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
void
BtreeIndex<Key,Compare,KeyTraits,PageSize>::next_iter(Iterator& it)
{
	Page* p = get_page(it._page);
	if (++it._slot < p->num_keys()) return;
	if (p->next_page() != size_t(-1)) {
		it._page = p->next_page();
		it._slot = 0;
	} else it = end();
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
void
BtreeIndex<Key,Compare,KeyTraits,PageSize>::update_link(size_t page, size_t slot, size_t link)
{
	Page* p = get_page(page);
	p->update_key(slot, link);
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::Page*
BtreeIndex<Key,Compare,KeyTraits,PageSize>::get_page(size_t pageId)
{
	if (pageId == size_t(-1)) return null;
	else if (pageId == 0) return _root;
	else return _pageCache.get_page(pageId);
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::Page*
BtreeIndex<Key,Compare,KeyTraits,PageSize>::get_free_page()
{
	Page* p = get_page(_root->next_free_page());
	if (p) {
		_root->rem_free_page(p);
	} else {
		_xstream.seekp(0, std::ios_base::end);
		size_t pageId = _xstream.tellp();
		p = new Page(pageId);
		p->encode(_xstream); // flush it so the next call works correctly
		_pageCache.add_page(p);
	}
	return p;
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
void
BtreeIndex<Key,Compare,KeyTraits,PageSize>::add_free_page(Page* p)
{
	_root->add_free_page(p);
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::iterator
BtreeIndex<Key,Compare,KeyTraits,PageSize>::lower_bound(Page* p, int depth, const Key& key)
{
	if (depth == 0) {
		Page::KeyList& kl = p->key_list();
		Page::KeyList::iterator it = kl.begin();
		size_t slot = 0;
		for (; it != kl.end(); ++it, ++slot) {
			if ( not _comp((*it).key, key) ) break;
		}
		if (it != kl.end()) {
			return Iterator(this, p->page_id(), slot);
		} else if (p->next_page() == size_t(-1)) {
			return end();
		} else {
			return Iterator(this, p->next_page(), 0);
		}
	} else {
		// find the next page to look on
		Page::KeyList& kl = p->key_list();
		Page::KeyList::iterator it = kl.begin();
		for (++it; it != kl.end(); ++it) {
			if ( not _comp((*it).key, key) ) break;
		}
		--it;
		return lower_bound(get_page((*it).link), depth-1, key);
	}
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::iterator
BtreeIndex<Key,Compare,KeyTraits,PageSize>::upper_bound(Page* p, int depth, const Key& key)
{
	if (depth == 0) {
		Page::KeyList& kl = p->key_list();
		Page::KeyList::iterator it = kl.begin();
		size_t slot = 0;
		for (; it != kl.end(); ++it, ++slot) {
			if ( _comp(key, (*it).key) ) break;
		}
		if (it != kl.end()) {
			return Iterator(this, p->page_id(), slot);
		} else if (p->next_page() == size_t(-1)) {
			return end();
		} else {
			return Iterator(this, p->next_page(), 0);
		}
	} else {
		// find the next page to look on
		Page::KeyList& kl = p->key_list();
		Page::KeyList::iterator it = kl.begin();
		for (++it; it != kl.end(); ++it) {
			if ( _comp(key, (*it).key) ) break;
		}
		--it;
		return upper_bound(get_page((*it).link), depth-1, key);
	}
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::iterator
BtreeIndex<Key,Compare,KeyTraits,PageSize>::insert(Page* p, int depth, const value_type& x)
{
	if (depth == 0) {	// leaf page
		size_t slot = p->add_key(*this, x);
		return Iterator(this, p->page_id(), slot);
	} else {			// internal page
		Iterator it;
		Page* nextp;
		Page::KeyList& kl = p->key_list();
		Page::KeyList::iterator kit = kl.begin();
		if ( _comp(x.first, (*kit).key) ) {
			// Boundary condition - key is less than first key on this page
			nextp = get_page((*kit).link);
			it = insert(nextp, depth-1, x);
			p->update_page_link(nextp);
		} else {
			// find which page to put the new key in
			for (++kit; kit != kl.end(); ++kit) {
				if ( _comp(x.first, (*kit).key) ) break;
			}
			--kit;
			nextp = get_page((*kit).link);
			it = insert(nextp, depth-1, x);
		}
		if (nextp->must_split()) {
			Page* newp = nextp->split(*this);
			p->insert_page_link(++kit, newp);
		}
		return it;
	}
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
std::pair<bool,bool>
BtreeIndex<Key,Compare,KeyTraits,PageSize>::erase(Page* p, int depth, const Key& key, Iterator pos)
{
	using std::pair;
	using std::make_pair;

	// we have to find the correct path to the page in pos
	if (depth == 0) {
		if (pos._page != p->page_id()) return std::make_pair(false, false);
		// Ok - remove the key
		bool notHalfFull = p->remove_key(pos._slot);
		return std::make_pair(true, notHalfFull);
	} else {
		// find the link we think will lead to pos
		Page::KeyList& kl = p->key_list();
		Page::KeyList::iterator kit = kl.begin();
		for (++kit; kit != kl.end(); ++kit) {
			if (not _comp(key, (*kit).key)) break;	// use lower bound test
		}
		--kit;	// back up one link
		Page* nextp = get_page((*kit).link);
		pair<bool,bool> rtn = make_pair(false, false);
		for (;;) {
			rtn = erase(nextp, depth-1, key, pos);
			if (rtn.first) break;
			++kit;
			if (kit == kl.end()) return make_pair(false, false);
			nextp = get_page((*kit).link);
		}
		
		// if rtn.first we got to the right page and updated it
		// now see what we have to do at this level
		if (nextp->num_keys() == 0) {
			add_free_page(nextp);
			bool notHalfFull = p->remove_page_link(nextp);
			return make_pair(true, notHalfFull);
		}
		// See if the page link is still correct
		if ( _comp((*kit).key, nextp->key_list().front().key) ) {
			// we removed or updated the first key in the lower page
			p->update_page_link(nextp); // might grow the page
		}
		if (rtn.second) {
			// the lower page is less than half full
			if (_dStrategy != None) {
				Page* oldp = nextp->consolidate(*this);
				if (oldp) {
					bool notHalfFull = p->remove_page_link(oldp);
					return make_pair(true, notHalfFull);
				}
				if (_dStrategy == Shuffle) {
					Page* op = nextp->shuffle(*this);
					p->update_page_link(op);	// might grow the page
				}
			}
		} else if (nextp->must_split()) {
			// the lower page might have grown
			Page* newp = nextp->split(*this);
			p->insert_page_link(++kit, newp);
		}
		return make_pair(true, false);
	}
}


template<typename Key, typename Compare, typename KeyTraits, size_t PageSize>
void
BtreeIndex<Key,Compare,KeyTraits,PageSize>::flush(UpdateStrategy s)
{
	_pageCache.reset();	// remove excess pages
	if (s == Safe) {
		_pageCache.flush();
		if (_root->is_total_dirty() or _root->is_dirty())
			_root->encode(_xstream);
	} else if (s == MostlySafe) {
		_pageCache.flush();
		if (_root->is_dirty()) _root->encode(_xstream);
	}
}


} //> namespace
