/*	Copyright (C) 2001,2002,2003 Ronnie Maor and Michael Brand
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * As a special exception, the copyright holders give permission
 * for additional uses of the text contained in its release of XPARAM.
 *
 * The exception is that, if you link the XPARAM library with other files
 * to produce an executable, this does not by itself cause the
 * resulting executable to be covered by the GNU General Public License.
 * Your use of that executable is in no way restricted on account of
 * linking the XPARAM library code into it.
 *
 * This exception does not however invalidate any other reasons why
 * the executable file might be covered by the GNU General Public License.
 *
 * If you write modifications of your own for XPARAM, it is your choice
 * whether to permit this exception to apply to your modifications.
 * If you do not wish that, delete this exception notice.  
 */

header "post_include_hpp" {

#include "../xparam/xp_value_management.h"
#include "xp_parser.h"

}


header "post_include_cpp" {
#include "../xparam/xp_error.h"

using namespace antlr;
using namespace std;
}

options {
  language = "Cpp";
  namespace = "xParam_internal";
  namespaceStd="std";
  namespaceAntlr="antlr";     
}

class xParamParser extends Parser;
options {
  defaultErrorHandler = false;
}
{
  public:
    void set_context(const std::vector<std::string>& redirection_context);

  private:
    static std::string m_int_to_string(int i); 

    static char m_decode_char(const std::string& str);
    static std::string m_decode_string(const std::string& str);

    static Handle<Value> m_create_value(const std::string& type_name, const ValueList& args);
    static Handle<Value> m_create_tentative_value(const std::string& encoded_value);
    static Handle<Value> m_constant_or_tentative_string(const std::string& token);

    std::string m_convert_path(const std::string& path) const;
    Handle<std::istream> m_open_url(const std::string& url_name) const;
    std::vector<std::string> m_new_context(const std::string& url_name) const;

    Handle<Value> m_url_read_value(const std::string& url_name) const;
    void m_url_read_value_set(const std::string& url_name, AssignmentListener& listener) const;
    
    std::vector<std::string> m_redirection_context;
}

////////////////////////////////////////////////////////////////
// public parser symbols
////////////////////////////////////////////////////////////////

value_set
[ AssignmentListener& listener ]
	:
	( value_set_member[listener] )* (";" | EOF)
;
exception
	catch [ANTLRException& ex] {
		throw Error(ex.what());
	}

value
returns [Handle<Value> val]
	:
	  val = basic_type
	| val = compound_value
	| id:ID {val=m_constant_or_tentative_string(id->getText());}
	| val = list_value
	| url:URL {val=m_url_read_value(url->getText());}
;
exception
	catch [NoViableAltException& ex] {
		string msg;
		if (LT(1)->getType() == Token::EOF_TYPE)
			msg = "unexpected EOF";
		else
			msg = "unexpected token type";
		throw Error(msg);
	}
	catch [ANTLRException& ex] {
		throw Error(ex.what());
	}

////////////////////////////////////////////////////////////////
// value_set implementation
////////////////////////////////////////////////////////////////

protected value_set_member
[ AssignmentListener& listener ]
{
  Handle<Value> val;
}
	:
	  (ID "=") => member_assignment[listener]
	| (URL) => url:URL {m_url_read_value_set(url->getText(),listener);}
	| flag_setting[listener]
;

protected member_assignment
[AssignmentListener& listener]
{
  Handle<Value> val;
}
	:
	id:ID "=" val=global_scope_value {listener.notify(id->getText(),val);}
;
exception
	catch [Error& ex] {
		throw Error("while parsing assignment to parameter '" 
				+ id->getText() + "':\n" + ex.what());
	}
	catch [RecognitionException& ex] {
		throw Error("syntax error parsing assignment to parameter '"
				+ id->getText() + "'");
	}

protected global_scope_value
returns [Handle<Value> val]
{
  ValueList values;
}
	:
	  values = non_empty_list_of_values {
		if (values.size() == 1) { // it's a single value
			val = values[0];
		}
		else { // it's a list of values
			val = make_value_copy(values);
		}
	  }
;

protected flag_setting
[AssignmentListener& listener]
	:
	  flag:FLAG {
		string str = flag->getText();
		bool flag_value = (str[0] == '+');
		string flag_name = str.substr(1,str.size()-1);
		listener.notify(flag_name,make_value_copy(flag_value));
	  }
;

////////////////////////////////////////////////////////////////
// value implementation
////////////////////////////////////////////////////////////////

protected compound_value
returns [Handle<Value> val]
{
  ValueList args;
}
	:
	  tn:TYPENAME "(" args=list_of_values ")" 
		{val=m_create_value(tn->getText(),args);}
;
exception
	catch [Error& ex] {
		throw Error("while parsing constructor for '" + tn->getText() + "':\n" + ex.what());
	}
	catch [RecognitionException& ex] {
		throw Error("syntax error while parsing constructor for '" + tn->getText() + "'");
	}

protected list_value
returns [Handle<Value> val]
{
  ValueList values;
}
	:
	"[" values=list_of_values "]" { val = make_value_copy(values); }
;
exception
	catch [Error& ex] {
		throw Error("while reading list:\n" + string(ex.what()));
	}
	catch [RecognitionException& ex] {
		throw Error("syntax error while reading list");
	}


////////////////////////////////////////////////////////////////
// list_of_values implementation
////////////////////////////////////////////////////////////////

protected list_of_values
returns [ValueList list]
	:
	  // empty list
	| list=non_empty_list_of_values
;

protected non_empty_list_of_values
returns [ValueList list]
{
  Handle<Value> val;
}
	:
	  val=value {list.push_back(val);}
	  ("," val=value {list.push_back(val);})*
;
exception
	catch [Error& ex] {
		if (!list.empty())
			throw Error("while parsing element " 
					+ m_int_to_string(list.size())
					+ " of value list:\n" + ex.what());
		else
			throw;
	}
	catch [RecognitionException& ex] {
		throw Error("syntax error while parsing element "
				+ m_int_to_string(list.size())
				+ " of value list:\n" + ex.what());
	}

////////////////////////////////////////////////////////////////
// basic_type implementation
////////////////////////////////////////////////////////////////

protected basic_type
returns [Handle<Value> val]
	:
	  t:TENTATIVE { val=m_create_tentative_value(t->getText()); }
	| c:CHAR {
		char decoded_char = m_decode_char(c->getText());
		val = make_value_copy(decoded_char);
	  }
	| s:STRING {
		string decoded_string = m_decode_string(s->getText());
		val = make_value_copy(decoded_string);
	  }
;
