//  Boost Scoped pointer demonstration program
//
//  Copyright 2002 Jim Hyslop, all rights reserved
// 
//  Permission to copy, use, modify, sell and distribute this software
//  is granted provided this copyright notice appears in all copies.
//  This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//
// This software has been tested with version 1.28.0 of the Boost
// library (released May 16, 2002). See www.boost.org to download Boost.
// 
//

#include <boost/scoped_ptr.hpp>
#include <iostream>

using namespace boost;
using namespace std;

///////////////// README README README README ///////////////////////////
// Demo #1 - this file will not compile "out of the box". This is
// deliberate, to demonstrate scoped_ptr's ability to cause an
// error if you try to copy it. To allow it to compile, uncomment the
// following definition:
//
// #define USER_DEFINED_COPYING

class ScopedPtrDemo
{
    struct ScopedPtrDemoImpl;
    scoped_ptr<ScopedPtrDemoImpl> impl_;
public:
    ScopedPtrDemo( int initialValue );

#ifdef USER_DEFINED_COPYING
    ScopedPtrDemo( const ScopedPtrDemo &other );
    ScopedPtrDemo & operator = ( const ScopedPtrDemo & );
#endif

    void increment();
    void show( ostream & ) const;
    void reset();
};

ostream & operator << (ostream & o, const ScopedPtrDemo &demo)
{
    demo.show(o);
    return o;
}

struct ScopedPtrDemo::ScopedPtrDemoImpl
{
    int value_;
    ScopedPtrDemoImpl( int initial ) : value_(initial) 
    {
        cout << "Creating impl_" << endl;
    }
    ScopedPtrDemoImpl( const ScopedPtrDemoImpl &other )
    :
    value_(other.value_)
    {
        cout << "Creating (copying) impl_" << endl;
    }
    ~ScopedPtrDemoImpl()
    {
        cout << "impl_ destroyed" << endl;
    }

};

ScopedPtrDemo::ScopedPtrDemo( int initialValue )
: impl_(new ScopedPtrDemoImpl(initialValue))
{
}

void ScopedPtrDemo::increment()
{
    // Demonstrate the scoped_ptr::get() function:
    ScopedPtrDemoImpl * myImplementation = impl_.get();
    ++ myImplementation->value_;
}

void ScopedPtrDemo::show( ostream & out ) const
{
    // Demonstrate scoped_ptr::operator->
    out << "value: " << impl_ -> value_;
}

void ScopedPtrDemo::reset()
{
    // Demonstrate reset() function
    impl_.reset(0);
}

#ifdef USER_DEFINED_COPYING

// Demonstrates scoped_ptr::operator *
ScopedPtrDemo::ScopedPtrDemo( const ScopedPtrDemo &other )
:
impl_(new ScopedPtrDemoImpl(*other.impl_))
{
}

ScopedPtrDemo & ScopedPtrDemo::operator = ( const ScopedPtrDemo &other )
{
    cout << "Inside assignment operator... ";

    // This local scope shows more clearly the lifetime of tmp.
    {
        scoped_ptr<ScopedPtrDemoImpl> tmp( new ScopedPtrDemoImpl(*other.impl_));

        // the following line could also be written swap(tmp, impl_)
        impl_.swap(tmp);
    }
    cout << "Assignment complete" << endl;

    return *this;
}
#endif

void demoScopedPtr()
{
    ScopedPtrDemo demo1(1);
    cout << demo1 << " (should be 1)" << endl;

    // Important - see the README comments at the top of this file
    ScopedPtrDemo demo2(demo1);
    demo1.increment();
    cout << demo1 << " (should be 2)" << endl;
    cout << demo2 << " (should be 1)" << endl;

    ScopedPtrDemo demo3(0);
    cout << demo3 << " (should be 0)" << endl;

    demo3 = demo2;
    cout << demo3 << " (should be 1)" << endl;

    cout << "Resetting demo3, you should see a single destruction message... \n";
    demo3.reset();
    cout << "\ndemo3 is now reset.\nYou should now see two destruction messages as \n"
        "demo1 and demo2 go out of scope" << endl;
}

// Demo #2
// Just when you thought it was going to compile :=) A quick demo to show
// how scoped_ptr behaves when given an incomplete type. To enable the
// program to compile, uncomment the following definition:
// 
// #define SUPPRESS_INCOMPLETE_TYPE_DEMO

#ifndef SUPPRESS_INCOMPLETE_TYPE_DEMO
class Incomplete;
Incomplete * createIncomplete();

void demoIncompleteType()
{
    scoped_ptr<Incomplete> incomplete(createIncomplete());

    // The following line will compile. Some compilers may warn
    // you about it, but not all will.
    auto_ptr<Incomplete> incomplete(createIncomplete());
}
#endif